@@ -74,6 +74,7 @@ fn roxido_fn(options: Vec<NestedMeta>, item_fn: syn::ItemFn) -> TokenStream {
     };
     let mut longjmp = true;
     let mut invisible = false;
+    let mut module = String::new();
     for meta in options {
         let meta = meta.0;
         let meta_string = quote!(#meta).to_string();
@@ -98,10 +99,13 @@ fn roxido_fn(options: Vec<NestedMeta>, item_fn: syn::ItemFn) -> TokenStream {
                             panic!("Unsupported value '{value_string}' for {name_string}.")
                         }
                     },
-                    _ => panic!("Unsupported option '{name_string}'."),
+                    "module" => {
+                        module = value_string;
+                    }
+                    _ => panic!("Unsupported token '{name_string}'."),
                 }
             }
-            _ => panic!("Unsupported option '{meta_string}'."),
+            _ => panic!("Unsupported token '{meta_string}'."),
         }
     }
     let name = item_fn.sig.ident;
@@ -436,6 +440,10 @@ fn roxido_fn(options: Vec<NestedMeta>, item_fn: syn::ItemFn) -> TokenStream {
                 let filename = "roxido.txt";
                 if let Ok(mut file) = OpenOptions::new().append(true).create(true).open(filename) {
                     let mut line = String::new();
+                    if !module.is_empty() {
+                        line.push_str(module.as_str());
+                        line.push_str("::");
+                    }
                     line.push_str(&func_name);
                     for arg in arg_names.iter() {
                         line.push_str(", ");
@@ -459,8 +467,9 @@ fn roxido_fn(options: Vec<NestedMeta>, item_fn: syn::ItemFn) -> TokenStream {
         // leaks.  See https://docs.rs/crate/setjmp for background information.
         TokenStream::from(quote! {
             #[allow(clippy::useless_transmute)]
+            #[allow(clippy::not_unsafe_ptr_arg_deref)]
             #[no_mangle]
-            extern "C" fn #name(#new_args) -> SEXP {
+            pub extern "C" fn #name(#new_args) -> SEXP {
                 let result = std::panic::catch_unwind(std::panic::AssertUnwindSafe(|| {
                     let pc = &mut Pc::__private_new();
                     #( #generated_statements )*
@@ -498,8 +507,9 @@ fn roxido_fn(options: Vec<NestedMeta>, item_fn: syn::ItemFn) -> TokenStream {
     } else {
         TokenStream::from(quote! {
             #[allow(clippy::useless_transmute)]
+            #[allow(clippy::not_unsafe_ptr_arg_deref)]
             #[no_mangle]
-            extern "C" fn #name(#new_args) -> SEXP {
+            pub extern "C" fn #name(#new_args) -> SEXP {
                 let result: Result<SEXP, _> = std::panic::catch_unwind(std::panic::AssertUnwindSafe(|| {
                     let pc = &mut Pc::__private_new();
                     #( #generated_statements )*
